#!/usr/bin/python
" Exceptions that might be encountered within the Bombardier Client"
# BSD License
# Copyright (c) 2009, Peter Banka et al
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the GE Security nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

class InvalidConfigData(Exception):
    "Thrown if we are given bad configuration info."
    def __init__(self, name):
        Exception.__init__(self)
        self.name = name
    def __repr__(self):
        return "Package %s requires a reboot. Stopping..." % self.name
    def __str__(self):
        return self.__repr__()

class RebootRequiredException(Exception):
    "Thrown after a package is installed that notes a reboot is required."
    def __init__(self, name):
        Exception.__init__(self)
        self.name = name
    def __repr__(self):
        return "Package %s requires a reboot. Stopping..." % self.name
    def __str__(self):
        return self.__repr__()

class FeatureRemovedException(Exception):
    "Thrown when attempting to run an invalid action"
    def __init__(self, feature_name):
        Exception.__init__(self)
        self.feature_name  = feature_name
    def __repr__(self):
        return "Feature %s is obsolete" % (self.feature_name)
    def __str__(self):
        return self.__repr__()

class AbstractClassException(Exception):
    "Thrown when the Spkg superclass hasn't been overridden"
    def __init__(self, method_name):
        Exception.__init__(self)
        self.class_name  = method_name
    def __repr__(self):
        msg = "Method %s was attempted in an abstract class"
        return msg % (self.class_name)
    def __str__(self):
        return self.__repr__()

class DependencyLoopException(Exception):
    "Thrown when a package A depends on package B which depends on package A..."
    def __init__(self, chain):
        Exception.__init__(self)
        self.chain  = chain
    def __repr__(self):
        output  = "Package chain has a recursive loop in dependencies: "
        for pkn in self.chain:
            output += pkn + ','
        return output[:-1]
    def __str__(self):
        return self.__repr__()

class BadPackage(Exception):
    "Thrown when a package is not set up correctly."
    def __init__(self, pkn, errmsg):
        Exception.__init__(self)
        self.errmsg      = errmsg
        self.pkn = pkn
    def __repr__(self):
        return "%s: %s" % (self.pkn, self.errmsg)
    def __str__(self):
        return self.__repr__()



#!/usr/bin/env python
"""module is meant to be run on a linux machine. This is the method the server
uses to interact with a box: find out what pacakges need to be installed, etc.
Currently supports RHEL5 and Ubuntu Hardy. Other POSIX environments could
be easily added, including cygwin"""

# Copyright (C) 2005 Peter Banka

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import sys, optparse, StringIO, traceback, yaml, time, re

from bombardier_core.Cipher import Cipher
from bombardier_core.Logger import Logger
from bombardier_core.Config import Config
from bombardier_client.Repository import Repository
from bombardier_core.Exceptions import ConfigurationException
from bombardier_client.PackageV4 import PackageV4
from bombardier_client.PackageV5 import PackageV5
from bombardier_client.BombardierClass import Bombardier
from bombardier_core.mini_utility import get_progress_path, get_spkg_path
from bombardier_core.static_data import FIX, CHECK_STATUS, CONFIGURE, RECONCILE
from bombardier_core.static_data import VERIFY, INSTALL, UNINSTALL, PURGE
from bombardier_core.static_data import DRY_RUN, INIT, EXECUTE
from bombardier_core.static_data import OK, FAIL
from bombardier_core import CORE_VERSION
from bombardier_client import CLIENT_VERSION
import os
import base64
import zlib

STREAM_BLOCK_SIZE = 77

USAGE  = ["usage: %prog { -s | -r | -n } INSTANCE | "]
USAGE += ["       %prog { -c | -v | -i | -u } INSTANCE PACKAGE-NAME [ PACKAGE-NAME2 ... ] | "]
USAGE += ["       %prog { -f | -p } INSTANCE FULL-PACKAGE-NAME [FULL-PACKAGE-NAME2 ... ] | "]
USAGE += ["       %prog -x INSTANCE PACKAGE-NAME SCRIPT-NAME "]
USAGE += ['']
USAGE += ["INSTANCE := A complete bombardier client configuration"]
USAGE += ["PACKAGE-NAME := An individual software module"]
USAGE += ["FULL-PACKAGE-NAME := {PACKAGE-NAME}-{PACKAGE-REVISION}"]
USAGE += ["SCRIPT-NAME := The name of a maintenance script that resides in the package"]

class NoInstanceError(Exception):
    "Asking to work on an instance that doesn't exist"
    def __init__(self, instance_name):
        Exception.__init__(self)
        self.instance_name = instance_name
    def __repr__(self):
        msg = "Attempting operation on a non-existant instance: %s"
        msg = msg  % self.instance_name
        return msg

def exit_with_return_code(value):
    "Provide proper output to the CNM server"
    if type(value) != type(0):
        Logger.error("Invalid exit code, not an integer: %s" % value)
        value = FAIL
    Logger.warning("==EXIT-CODE==:%s" % value)
    sys.exit(value)

def find_likely_pkn(instance_name, pkn):
    """Sometimes an ambiguous package name is requested.
    This attempts to help a guy out."""
    status_yml = yaml.load(open(get_progress_path(instance_name)).read())
    pkns = []
    status_packages = status_yml['install-progress']
    for name in status_packages:
        if status_packages[name]['INSTALLED'] in [ 'NA', 'BROKEN' ]:
            continue
        if pkn.lower() in name.lower():
            pkns.append(name)
    if len(pkns) > 1:
        msg = 'Ambiguous package name: %s could be any of %s' % (pkn, str(pkns))
        Logger.error( msg )
        exit_with_return_code(FAIL)
    if len(pkns) == 0:
        Logger.error( 'Package not found: %s' %pkn )
        exit_with_return_code(FAIL)
    else:
        pkn = '-'.join(pkns[0].split('-')[:-1])
        Logger.info( 'Using %s' %pkn)
        return pkn

def fix_spkg(instance_name, pkn, action, package_factory):
    """A user is requesting to take a package that is currently flagged
    as 'broken' and set it to be 'installed'. Perhaps they have manually
    fixed the package on the system."""
    status_data = open(get_progress_path(instance_name), 'r').read()
    status = yaml.load(status_data)
    if status.get("install-progress") == None:
        status["install-progress"] = {}
        Logger.warning( "Status file is empty." )
    now = time.asctime()
    if action == FIX:
        fix_name = []
        base_names = re.compile("(\s+)\-\d+").findall(pkn)
        if base_names:
            base_name = base_names[0]
        else:
            base_name = pkn
        for possible_pkn in status["install-progress"]:
            if base_name in possible_pkn:
                fix_name.append(possible_pkn)
        if len(fix_name) > 1:
            msg = "Package name %s is ambigious. (possible %s)"
            msg = msg % (pkn, ' '.join(fix_name))
            Logger.error(msg)
            return FAIL
        elif len(fix_name) == 1:
            pkn = fix_name[0]
        elif len(fix_name) == 0:
            new_package = package_factory.get_me_one(pkn)
            pkn = new_package.full_name
            Logger.info("Selecting previously UNINSTALLED package: %s" % pkn)
        status["install-progress"]["%s" % pkn] = {"INSTALLED": now,
                                                  "UNINSTALLED": "NA",
                                                  "VERIFIED": now}
        Logger.info("==OUTPUT==:%s has been set to INSTALLED." % pkn )
    elif action == PURGE:
        if status["install-progress"].get(pkn):
            del status["install-progress"][pkn]
            msg = "==OUTPUT==:%s has been removed from %s status"
            msg = msg % (pkn, instance_name)
            Logger.info(msg)
        else:
            Logger.info("==OUTPUT==:%s is not in the status file" % pkn)
            pkns = status["install-progress"]
            possible_names = [x for x in pkns if pkn in x]
            msg = "==OUTPUT==:Maybe you want one of these: %s"
            msg = msg % str(possible_names)
            Logger.info(msg)
            return FAIL
    open(get_progress_path(instance_name), 'w').write(yaml.dump(status))
    return OK

class BombardierEnvironment:
    """Holds information about the system Configuration. Should
    probably be removed"""
    def __init__(self, instance_name):
        """instance_name -- a machine can have several different names
                            that it is referred to to that packages can
                            be installed more than once"""
        self.repository      = None
        self.config          = None
        self.instance_name   = instance_name

    def data_request(self):
        "Obtain configuration data from the server"
        b64_data = []
        while True:
            chunk = sys.stdin.read(STREAM_BLOCK_SIZE)
            if not chunk or chunk[0] == ' ':
                break
            b64_data.append(chunk)
        yaml_data = ''
        yaml_data = zlib.decompress(base64.decodestring(''.join(b64_data)))
        Logger.debug("Received %s lines of yaml" % len(yaml_data.split('\n')))

        try:
            input_data = yaml.load(yaml_data)
        except:
            ermsg = "Configuration data not YAML-parseable: %s" % (repr(yaml_data))
            raise ConfigurationException(ermsg)
        if type(input_data) == type("string"):
            ermsg = "Configuration data not YAML-parseable: %s" % (repr(yaml_data))
            raise ConfigurationException(ermsg)
        if type(input_data) != type({}) and type(input_data) != type([]):
            input_data = input_data.next()
        config_key = input_data.get("config_key", None)
        if config_key:
            enc_yaml_file = os.path.join(get_spkg_path(), self.instance_name,
                                         'client.yml.enc')
            if not os.path.isfile(enc_yaml_file):
                msg = "%s file doesn't exist" % enc_yaml_file
                raise ConfigurationException(msg)
            enc_data = open(enc_yaml_file).read()
            cipher = Cipher(config_key)
            plain_yaml_str = cipher.decrypt_string(enc_data)
            try:
                input_data = yaml.load(plain_yaml_str)
            except:
                ermsg = "Received bad YAML file: %s" % enc_yaml_file
                raise ConfigurationException(ermsg)

        config_data  = input_data.get("config_data")
        if not config_data:
            raise ConfigurationException("No configuration data received")
        package_data = input_data.get("package_data", {})
        self.config = Config(self.instance_name, config_data)
        self.repository = Repository(self.instance_name, package_data)

class PackageFactory:
    "Generate type-4 or type-5 packages"
    def __init__(self, env):
        self.env = env

    def get_me_one(self, pkn):
        '''
        create a package object based on pkn
        pkn -- the name of the package to create
        '''
        version = self.env.repository.determine_pkg_version(pkn)
        if version == 4:
            new_package = PackageV4(pkn, self.env.repository,
                                    self.env.config,
                                    self.env.instance_name)
        else:
            new_package = PackageV5(pkn, self.env.repository,
                                    self.env.config,
                                    self.env.instance_name)
        new_package.initialize()
        return new_package

def get_bc(instance_name, env):
    "This class is the primary interface to the rest of the system"
    bc_obj = Bombardier(env.repository, env.config, instance_name)
    return bc_obj

def instance_setup(instance_name):
    "When Bombardier is first getting set up, this method is called."
    progress_path = get_progress_path(instance_name)
    status_dict = None
    if os.path.isfile(progress_path):
        try:
            status_dict = yaml.load(open(progress_path).read())
        except:
            msg = "Unable to load existing yaml from %s" % progress_path
            Logger.warning(msg)
    if type(status_dict) != type({}):
        status_dict = {"install-progress": {}}
    status_dict["client_version"] = CLIENT_VERSION
    status_dict["core_version"] = CORE_VERSION
    status_dict["clientVersion"] = CLIENT_VERSION
    status_dict["coreVersion"] = CORE_VERSION
    pkg_dir = os.path.join(get_spkg_path(), instance_name, "packages")
    if not os.path.isdir(pkg_dir):
        os.makedirs(pkg_dir)
    open(progress_path, 'w').write(yaml.dump(status_dict))

def process_action(action, instance_name, pkn, method_name,
                   package_factory):
    """
    Performs a Bombardier action on this system
    action -- either INIT, INSTALL, UNINSTALL, VERIFY, CONFIGURE,
              EXECUTE, RECONCILE, CHECK_STATUS, FIX, or PURGE
    instance_name -- the name of the machine
    pkn -- the package name to operate on. Could be none for actions
           that operate on the whole machine, such as RECONCILE,
           CHECK_STATUS, or INIT 
    method_name -- Only applicable for the EXECUTE action, runs a named
                   method within a package.
    package_factory -- provides packages we may need
    """
    if action == INIT:
        instance_setup(instance_name)
        return OK

    if action in [ UNINSTALL, VERIFY, CONFIGURE, EXECUTE ]:
        pkn = find_likely_pkn(instance_name, pkn)

    status = FAIL
    try:
        if action in [ FIX, PURGE ]:
            status = fix_spkg(instance_name, pkn, action, package_factory)
            return status
        bc_obj = get_bc(instance_name, package_factory.env)
        if action == CHECK_STATUS:
            status_dict = bc_obj.check_system()
            if type(status_dict) == type({}):
                if status_dict["broken"]:
                    Logger.info("BROKEN PACKAGES:")
                    for pkn in status_dict["broken"]:
                        Logger.info("- %s" % pkn)
                status = OK
            else:
                status = FAIL
        elif action in [ RECONCILE, DRY_RUN ]:
            bc_obj.record_errors = True
            status = bc_obj.reconcile_system(action)
        else:
            bc_obj.record_errors = False
            status = bc_obj.use_pkg(pkn, action, method_name)
    except:
        err = StringIO.StringIO()
        traceback.print_exc(file = err)
        err.seek(0)
        data = err.read()
        ermsg = ''
        for line in data.split('\n'):
            ermsg += "\n||>>>%s" % line
        Logger.error(ermsg)
        return FAIL
    return status

def main():
    Logger.add_std_err_logging()

    parser = optparse.OptionParser('\n'.join(USAGE))

    parser.add_option("-s", "--status", dest="action",
                      action="store_const", const=CHECK_STATUS,
                      help="display the status of the system")
    parser.add_option("-c", "--configure", dest="action",
                      action="store_const", const=CONFIGURE,
                      help="configure a package")
    parser.add_option("-v", "--verify", dest="action",
                      action="store_const", const=VERIFY,
                      help="verify a package")
    parser.add_option("-i", "--install", dest="action",
                      action="store_const", const=INSTALL,
                      help="install a package")
    parser.add_option("-r", "--reconcile", dest="action",
                      action="store_const", const=RECONCILE,
                      help="reconcile the system")
    parser.add_option("-x", "--execute", dest="action",
                      action="store_const", const=EXECUTE,
                      help="Execute a maintenance script")
    parser.add_option("-u", "--uninstall", dest="action",
                      action="store_const", const=UNINSTALL,
                      help="uninstall a package")
    parser.add_option("-f", "--fix", dest="action",
                      action="store_const", const=FIX,
                      help="set a package status to INSTALLED without doing anything")
    parser.add_option("-p", "--purge", dest="action",
                      action="store_const", const=PURGE,
                      help="Remove a package from the status")
    parser.add_option("-d", "--dry-run", dest="action",
                      action="store_const", const=DRY_RUN,
                      help="Do a test reconcile")
    parser.add_option("-n", "--init", dest="action",
                      action="store_const", const=INIT,
                      help="Initialize the client after installation")

    (options, args) = parser.parse_args()
    if len(args) < 1:
        print "CMD: %s" % ' '.join(sys.argv)
        print "This command requires an instance name."
        parser.print_help()
        exit_with_return_code(1)
    instance_name = args[0]

    env = BombardierEnvironment(instance_name)
    env.data_request()
    package_factory = PackageFactory(env)

    if options.action in [ RECONCILE, CHECK_STATUS, DRY_RUN, INIT ]:
        status = process_action(options.action, instance_name,
                               '', '', package_factory)
    else:
        method_name   = ""
        if len(args) < 2:
            print "CMD: %s" % ' '.join(sys.argv)
            print "This command requires a package name as an argument."
            parser.print_help()
            exit_with_return_code( 1 )
        pkns = args[1:]
        if options.action == EXECUTE:
            if len(args) != 3:
                print "CMD: %s" % ' '.join(sys.argv)
                print "This command requires a package name and a script name."
                parser.print_help()
                exit_with_return_code( 1 )
            pkns = [args[1]]
            method_name = args[2]

        for pkn in pkns:
            status = process_action(options.action, instance_name,
                                    pkn, method_name,
                                    package_factory)
            if status != OK:
                exit_with_return_code(status)
    exit_with_return_code(status)

if __name__ == "__main__":
    main()



#!/usr/bin/python
# BSD License
# Copyright (c) 2009, Peter Banka et al
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the GE Security nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''Contains those classes necessary for figuring out package installation
order and offers major actions available on a machine. Main entry
point for BC

NOTE ON NAMING CONVENTIONS: 
 - 'pkn' = a package name
 - 'pkd' = a package dictionary
 - 'pkg' = a package object
 - 'vpn' = "virtual package name"
 - 'pkl' = a package list
 - 'pdat' = a dictionary of package progress data
'''

import sets, os, copy
import PackageV4, PackageV5
import Exceptions
from bombardier_core.Logger import Logger
from bombardier_core.mini_utility import get_installed, get_progress_data
from bombardier_core.mini_utility import diff_dicts
from bombardier_core.static_data import OK, FAIL, HASH_FILE
from bombardier_core.static_data import UNINSTALL, DRY_RUN, VERIFY
from bombardier_core.static_data import INSTALL, CONFIGURE
from bombardier_core.static_data import EXECUTE, ACTION_REVERSE_LOOKUP

MAX_CHAIN_DEPTH = 50

def swap(list_obj, index1, index2):
    'swap two indexes in a list'
    new_list = copy.deepcopy(list_obj)
    new_list[index1] = list_obj[index2]
    new_list[index2] = list_obj[index1]
    return new_list

class PackageChain:
    ''' 
    A list of package names which should be installed in order. Contains
    the priority of the most important package in the chain
    '''
    def __init__(self, priority, start_pkn, pkd,
                 installed_pkns, broken_pkns, repository,
                 config, instance_name, record_errors):
        """
        priority -- How important this package chain is (higher is better)
        start_pkn -- The name of the first package in this 
                     dependency chain
        pkd -- a dictionary of package objects
        installed_pkns -- a list of packages that have been installed
        repository -- the Repository object which keeps track of packages
        config -- the object that tracks of the configuration fo the system
        instance_name -- the name of this machine (can be more than one)
        record_errors -- flag whether dependency errors should be tracked
        """
        self.depth       = 0
        self.priority    = priority
        self.pkd         = pkd
        self.instance_name = instance_name
        self.chain       = [start_pkn]
        self.repository  = repository
        self.config      = config
        self.record_errors = record_errors
        self.v_pkgs   = VirtualPackages(repository.pkg_data)
        self.installed_pkns = self.v_pkgs.resolve_vpkl( installed_pkns )
        self.broken_pkns = broken_pkns
        self._pkg_chain(start_pkn)
        self._filter_bad_pkgs()

    def _filter_bad_pkgs(self):
        '''
        Go through my chain of packages, and if there are any 'bad' packages
        in the list of dependencies, cut off the package chain at that point
        '''
        index = 0
        for index in range(0, len(self.chain)):
            pkn = self.chain[index]
            if pkn in self.broken_pkns:
                index -= 1
                msg = "Omitting packages %s due to bad package"
                msg = msg % self.chain[index:]
                Logger.warning(msg)
                break
        self.chain = self.chain[:index+1]

    def _pkg_chain(self, pkn):
        '''
        Recursive function to add to self.chain packages in a proper
        dependency order
        pkn -- the name of a package or virtual package
        '''
        self.depth += 1
        if self.depth > MAX_CHAIN_DEPTH:
            raise Exceptions.DependencyLoopException(self.chain)
        for dep_name in self.pkd[ self.get_actual_pkn( pkn ) ].dependencies:
            if dep_name not in self.installed_pkns: 
                self.chain.insert( 0, self.get_actual_pkn( dep_name ) )
            else:
                continue
            self._sync_dependencies( dep_name, pkn )
            dependency = self.pkd.get(self.get_actual_pkn( dep_name ))
            if dep_name in self.broken_pkns:
                self.priority = 0
            else:
                self.priority = max( dependency.priority, self.priority )
            self._pkg_chain( dep_name )
        self.depth -= 1

    def _sync_dependencies( self, dep_name, pkn ):
        '''
        We have discovered that the 'pkn' package is dependent upon
        the 'dep_name' package being installed. Therefore, we can't install
        pkn unless dep_name is installed. We want to make sure that
        the dep_name package is going to be installed.
        dep_name -- the name of a package which is dependent upon one
                    that should be installed
        pkn -- the name of the package that dep_name is
               dependent upon
        '''
        if dep_name not in ( self.v_pkgs.resolve_vpkl( self.installed_pkns ) + \
                            self.v_pkgs.resolve_vpkl( self.pkd.keys() ) ):
            self.pkd[dep_name] = self._get_new_pkg( dep_name )
            if self.record_errors:
                self.pkd[pkn].add_dependency_error( dep_name )

    def get_actual_pkn( self, pkn ):
        '''
        This resolves names of packages into 'actual' package names,
        not virtual names
        pkn -- the name of either a package or virtual package
        '''
        return( self.v_pkgs.get_actual_pkn( pkn, self.pkd.keys() ) )

    def _get_new_pkg( self, pkn ):
        '''
        create a package object based on pkn
        pkn -- the name of the package to create
        '''
        version = self.repository.determine_pkg_version(pkn)
        pkg = None
        if version == 4:
            pkg = PackageV4.PackageV4(pkn, self.repository,
                                          self.config,
                                          self.instance_name)
        else:
            pkg = PackageV5.PackageV5(pkn, self.repository,
                                          self.config,
                                          self.instance_name)
        pkg.initialize()
        return pkg

class VirtualPackages:
    '''A lookup class that can resolve names of virtual package names to
    actual package names'''
    def __init__( self, pkd ):
        '''
        pkd -- a dictionary of package objects
        '''
        self.pkd = pkd
        self.virtual_pkgs_dict = {}
        self.virtual_name_lookup_dict = {}
        self._init_vpkg_dictionaries()

    def _init_vpkg_dictionaries( self ):
        '''
        initialize the dictionaries that this class uses to resolve virtual
        package names into actual package names
        '''
        for pkn in self.pkd:
            pkg_data = self.pkd[pkn]
            if pkg_data.get( 'virtualpackage' ):
                vpn = pkg_data['virtualpackage'] 
                self.virtual_name_lookup_dict.setdefault( pkn, vpn )
                self.virtual_pkgs_dict.setdefault( vpn, [] ).append( pkn )

    def _get_pkn_list_from_vpn( self, vpn ):
        '''
        We are given the name of a virual package. We are to find the names
        of all the actual packages which could be used in place of that
        virtual
        vpn -- the name of a virtual package
        '''
        return( self.virtual_pkgs_dict.get( vpn, [] ) )

    def get_vpn_from_pkn( self, pkn ):
        '''
        Return the name of an actual package. If the pkn submitted
        is an actual package, return it, otherwise look up an actual one and
        substitute it.
        pkn -- the name of either an actual package or virual
               package name
        '''
        return( self.virtual_name_lookup_dict.get( pkn, pkn ) )
    
    def resolve_vpkl( self, pkl ):
        '''
        go through a list of package names, replacing any virtual package names
        with actual package names
        pkl -- a list of package names, some of which could be
               'virtual', meaning that they represent a class of
               possible packages
        '''
        return( [self.get_vpn_from_pkn( pkn ) for pkn in pkl] )

    def get_actual_pkn( self, pkn, pkns ):
        '''
        We are given a pkn, which may be actual or virtual. We are
        going to find all the actual package names which could be used in place
        of that package name. We only want to pick a package which is in our
        repository.
        pkn -- an actual or virtual package name
        pkns -- a list of actual package names
        '''
        resolved_pkns = self._get_pkn_list_from_vpn( pkn )
        installable_pkns = []
        for actual_pkn in resolved_pkns:
            if actual_pkn in pkns:
                installable_pkns.append(actual_pkn)
        installable_pkns.append(pkn)
        return( installable_pkns[0] ) 

def find_differences(pkg_config, config_diff, differences=[], chain=[]):
    '''
    Find the differences in the configuration fo the system since the last
    time the package was installed or configured
    package_config -- the configuration values required by a package
    config_diff -- the differences that were made to this system since
                   the last time this package was installed.
    differences -- differences in the configuration that are relevant
                   to this package (or any other installed package on the
                   machine, as it recurses)
    chain -- a package chain
    '''
    output = differences
    for key in pkg_config.keys():
        if type(pkg_config[key]) == type({}):
            if key not in config_diff:
                continue
            _newdif = find_differences(pkg_config[key], config_diff[key],
                                     output, chain + [key])
            continue
        if key in config_diff.keys():
            output.append("%s/%s" % ('/'.join(chain), key))
            continue
    return output

class Bombardier:
    '''Master class for managing system actions'''

    def __init__(self, repository, config, instance_name):
        '''
        repository -- object that keeps track of package data
        config -- object that keeps track of this machine's configuration
        instance_name -- the name of this machine
        '''
        self.repository = repository
        self.config     = config
        self.instance_name    = instance_name
        self.record_errors = True
        self.operation_status = OK
        self.operation_output = []

    def _get_dependency_errors(self, bom_pkns, pdat):
        '''
        Pulls dependency error data out of status.yml
        bom_pkns -- the names of packages which are on the bill of
                    materials for this machine
        pdat -- a dictionary which maintains information about what
                packages are installed, uninstalled, verified, etc.
        '''
        dependency_names = set([])
        install_progress = pdat.get('install-progress', {})
        for pkn in install_progress:
            pkd = install_progress[ pkn ]
            dep_errors = set(pkd.get('DEPENDENCY_ERRORS', []))
            dependency_names = dependency_names.union(dep_errors)
        dependency_names = list(dependency_names - set(bom_pkns))
        return dependency_names

    def _create_pkg_chains(self, pkd):
        '''
        Create package chains based on current set of packages
        pkd -- dictionary of package objects
        '''
        chains = []
        pkg_data = get_progress_data(self.instance_name, False)
        installed_pkns, broken_pkns = get_installed(pkg_data)
        for pkn in pkd.keys():
            if pkn in broken_pkns:
                Logger.warning("Skipping broken package %s" % pkn)
                continue
            if pkn not in installed_pkns:
                chain_priority = pkd[pkn].priority
            else:
                chain_priority = 0
            try:
                new_chain = PackageChain(chain_priority, pkn,
                                         pkd, installed_pkns,
                                         broken_pkns, self.repository,
                                         self.config,
                                         self.instance_name, self.record_errors)
            except Exceptions.BadPackage, err:
                errmsg = "Package %s will not be installed because it is "\
                         "dependent upon one or more broken packages"
                Logger.warning(errmsg % pkn)
                Logger.warning(str(err))
                continue
            chains.append([new_chain.priority, new_chain.chain])
        return chains

    def _get_top_priority(self, chains):
        '''
        Go through a bunch of chains and determine which is the most important
        chains -- tuples of priority / chain
        '''
        top_priority = 0
        for priority, _chain in chains:
            if priority > top_priority:
                top_priority = priority
        return top_priority

    def _find_install_order(self, pkd):
        """Returns a list of pkns in the correct installation order
        - Create chains of package dependencies, the priority of each
        chain is the package with the highest priority in it, unless
        that package is already installed.
        - A package can appear in more than one chain
        - A package chain can have a single package in it if it does not
        have any dependencies and is not dependent on others.
        pkd -- dictionary of package objects
        """

        chains = self._create_pkg_chains(pkd)
        pdat = get_progress_data(self.instance_name, False)
        installed_pkns, _broken_pkns = get_installed(pdat)
        # - Put all the packages of each chain into the installation
        # order, excluding those that have already been installed in order
        # of chain priority. If a package is already in the installation
        # List, do not put it in twice.
        install_order = []
        while chains:
            top_priority = self._get_top_priority(chains)
            for priority, chain in chains:
                if priority == top_priority:
                    for pkn in chain:
                        if pkn not in installed_pkns:
                            if pkn not in install_order:
                                install_order.append(pkn)
                    chains.remove([priority, chain])
        return install_order

    def _install_pkgs(self, add_pkd, dry_run=False):
        '''
        Provided a dictionary of packages, we will install those that need to
        be installed.
        add_pkd -- a dictionary of packages that need to be added
        dry_run -- whether or not to just 'fake it'
        '''
        making_progress = True
        pkns_left = ['initialize']
        while making_progress and pkns_left:
            making_progress = False
            install_order = self._find_install_order(add_pkd)
            pkns_left = []
            map(pkns_left.append, install_order)
            for pkn in install_order:
                msg = "Packages remaining to install (in order):"
                for pkn in pkns_left:
                    Logger.info("  +   %s" % pkn)
                pkns_left.remove(pkn)
                pkg = add_pkd[pkn]
                erstr = "Currently installing package priority %s [%s]"
                Logger.info(erstr % (pkg.get_priority(), pkn))
                status = pkg.install_and_verify(pkns_left)
                if not dry_run:
                    hash_path = os.path.join(pkg.get_path(), HASH_FILE)
                    self.config.save_hash(hash_path)
                if status == FAIL:
                    erstr = "Package installation failure -- re-calculating"\
                            " package installation order"
                    Logger.error(erstr)
                    break
                else:
                    making_progress = True
        if pkns_left:
            msg = "There are packages that are broken, and we have done all"\
                  " we can do. ; ;"
            Logger.error(msg)
            return FAIL
        return OK

    def _cleanup(self):
        'Some housekeeping before exiting'
        if self.operation_status == FAIL:
            log_function = Logger.error
        else:
            log_function = Logger.info
        for message in self.operation_output:
            log_function("==OUTPUT==:%s" % message)
        return self.operation_status

    def _check_configuration(self, pkg):
        '''
        Check to see if we have sufficient configuration data to mess around
        with a given package
        pkg -- a package object that we need to do something with
        '''
        if pkg.get_configuration():
            required_configs = pkg.get_configuration()
            diff = diff_dicts(required_configs, self.config.data)
            if diff != {}:
                self.operation_status = FAIL
                errmsg = "This machine does not have sufficient "\
                         "configuration data to install %s " % pkg.name
                diff_txt = ''
                for key in diff:
                    diff_txt += "key: %s; value: %s" % (key, diff[key])
                Logger.warning(errmsg)
                Logger.warning(diff_txt)
                raise Exceptions.BadPackage(pkg.name, "Bad Config")

    def _get_new_pkg(self, pkn):
        '''
        Instantiate a package object based on name
        pkn -- The name of a package to instantiate
        '''
        version = self.repository.determine_pkg_version(pkn)
        pkg = None
        if version == 4:
            pkg = PackageV4.PackageV4(pkn, self.repository,
                                          self.config,
                                          self.instance_name)
        else:
            pkg = PackageV5.PackageV5(pkn, self.repository,
                                          self.config,
                                          self.instance_name)
        pkg.initialize()
        return pkg

    def _get_add_pkd(self, add_pkns):
        '''
        create a pkd for package names that should be installed
        add_pkns -- the names of packages that we should be installing
        '''
        pkd = {}
        for pkn in add_pkns:
            try:
                pkg = self._get_new_pkg(pkn)
                self._check_configuration(pkg)
                pkd[pkn] = pkg
            except Exceptions.BadPackage, err:
                errmsg = "Skipping bad package %s: %s" % (err.pkn, err.errmsg)
                self.operation_output.append(errmsg)
                Logger.warning(errmsg)
        return pkd

    def _get_del_pkd(self, del_pkns):
        '''
        create a dictionary of package objects based on a list of names
        that we know we're supposed to remove
        del_pkns -- the names of packages that we should be removing
        '''
        pkd = {}
        for pkn in del_pkns:
            try:
                pkg = self._get_new_pkg(pkn)
                pkg.action = UNINSTALL
                pkd[pkn] = pkg
            except Exceptions.BadPackage, err:
                errmsg = "Skipping Bad package: %s" % err
                Logger.warning(errmsg)
        return pkd

    def _sort_uninstalled_pkgs(self, uninstall_order):
        '''Make sure packages get uninstalled in the right order, paying
        attention to dependencies.
        uninstall_order -- a list of package names the order of which indicates
                           which should be uninstalled first, second...
        '''
        dependency_dict = {}
        if len(uninstall_order) > 1:
            Logger.info("Determining uninstallation order...")
            for pkn in uninstall_order:
                dependency_dict[pkn] = []

            for pkn in uninstall_order:
                pkg = self._get_new_pkg(pkn)
                for other_pkn in uninstall_order:
                    if other_pkn in pkg.dependencies:
                        dependency_dict[other_pkn].append(pkn)
        else:
            return uninstall_order

        new_proper_order = copy.deepcopy(uninstall_order)
        swapped = True

        while swapped:
            proper_order = copy.deepcopy(new_proper_order)
            swapped = False
            for pkn in proper_order:
                dependency_list = dependency_dict[pkn]
                for dependency in dependency_list:
                    index1 = proper_order.index(dependency)
                    index2 = proper_order.index(pkn)
                    if index1 > index2:
                        new_proper_order = swap(proper_order, index1, index2)
                        swapped = True
                        break
                if swapped:
                    break

        Logger.info("Uninstallation Order: %s" % new_proper_order)
        return new_proper_order

    def _get_uninst_pkg_dep(self, pkd,
                            del_pkns,
                            installed_pkns):
        '''Add any packages that are installed already which are dependent
        upon those to the list as well
        pkd -- The dictionary of package objects
        del_pkns -- names of packages we want to remove
        installed_pkns -- names of packages that are installed already
        '''
        msg = "Checking dependencies of packages to be uninstalled %s..."
        msg = msg % del_pkns
        Logger.info(msg)
        v_pkgs = VirtualPackages(self.repository.pkg_data)
        uninstall_order = copy.deepcopy(del_pkns)
        while del_pkns:
            new_dependency_names = []
            del_pkns = []
            for pkn in installed_pkns:
                if pkn in pkd.keys():
                    Logger.debug("Package %s will already be deleted -- "\
                                 "ignoring" % pkn)
                    continue
                pkg = self._get_new_pkg(pkn)
                pkg.action = UNINSTALL
                for tombstoned_pkns in pkd.keys():
                    vpn = v_pkgs.get_vpn_from_pkn(tombstoned_pkns)
                    if vpn in pkg.dependencies:
                        erstr = "Adding to package removal list: %s"\
                                " (depends on %s)"
                        uninstall_order.insert(0, pkn)
                        Logger.info(erstr % (pkn, tombstoned_pkns))
                        pkd[tombstoned_pkns].depends_on_me.append(pkn)
                        if pkn not in pkd.keys():
                            if pkn not in new_dependency_names:
                                pkd[pkn] = pkg
                                new_dependency_names.append(pkn)
                del_pkns = new_dependency_names

        proper_order = self._sort_uninstalled_pkgs(uninstall_order)
        return pkd, proper_order

    def _get_pkd_to_remove(self, del_pkns):
        '''
        Given a list of package names that we want to delete, let's create a
        proper dictionary of package objects which provide complete dependency
        information and specify the proper uninstallation order
        del_pkns -- a list of package names that need to be removed
        '''
        uninstall_order = []
        pdat = get_progress_data(self.instance_name, False)
        installed_pkns, _broken_pkns = get_installed(pdat)
        del_pkd = self._get_del_pkd(del_pkns)
        if sets.Set(installed_pkns) == sets.Set(del_pkd.keys()):
            return del_pkd, del_pkd.keys()
        if del_pkns:
            del_pkd, uninstall_order = self._get_uninst_pkg_dep(del_pkd,
                                                                del_pkns,
                                                                installed_pkns)
        return del_pkd, uninstall_order

    def _examine_dependencies(self, pkns):
        """ We need to determine if there are packages installed which
        have missing dependencies. If so, they should be added to the BOM.
        pkns -- a list of package names to check
        """
        updated_pkns = pkns.copy()
        vpkgs = VirtualPackages(self.repository.pkg_data)

        for pkn in pkns:
            apkn = vpkgs.get_actual_pkn(pkn, pkns)
            pkg = self._get_new_pkg(apkn)
            pkg.initialize()
            for dep in pkg.dependencies:
                updated_pkns.update([vpkgs.get_actual_pkn(dep, pkns)])
        if updated_pkns.difference(pkns):
            updated_pkns = self._examine_dependencies(updated_pkns)
        return updated_pkns
    
    def _check_bom(self, bom_pkns):
        """ Check through what should be installed on the system and what
        is installed on the system and determine what packages aren't
        installed that should be and what packages are installed that
        shouldn't be.
        bom_pkns -- A list of package names that are intended to be
                    installed on the machine (i.e. the 'bill of
                    materials'
        """
        should_be_installed = []
        shouldnt_be_installed = []
        pdat = get_progress_data(self.instance_name, False)
        installed_pkns, broken_pkns = get_installed(pdat)
        
        all_pkns = set(installed_pkns).union(broken_pkns)
        all_plus_missing_pkns = self._examine_dependencies(all_pkns)
        missing_pkns = list(all_plus_missing_pkns - all_pkns)
        bom_pkns = bom_pkns + missing_pkns

        dependency_errors = self._get_dependency_errors(bom_pkns, pdat)
        if dependency_errors:
            errmsg = "The following packages are installed as "\
                     "dependencies %s" % dependency_errors
            Logger.debug(errmsg)
        bom_pkns += dependency_errors
        for pkn in installed_pkns:
            if pkn not in bom_pkns:
                shouldnt_be_installed.append(pkn)
        for pkn in bom_pkns:
            if pkn not in installed_pkns:
                should_be_installed.append(pkn)
        return should_be_installed, shouldnt_be_installed

    def _ck_inst_stat(self, bom_pkns):
        '''
        Given a list of packages that are supposed to be on this machine, we
        will look at the installation status and (1) make a decision on what
        packages should be removed (and in what order) and (2) decide which
        packages should be added.
        bom_pkns -- A list of package names that are intended to be
                             installed on the machine (i.e. the 'bill of
                             materials'
        '''
        if bom_pkns == []:
            bom_pkns = self.config.get_bom_pkns()
            if bom_pkns == []:
                Logger.warning("Empty Bill of Materials")
        add_pkns, del_pkns = self._check_bom(bom_pkns)
        add_pkd = self._get_add_pkd(add_pkns)
        del_pkd, uninstall_order = self._get_pkd_to_remove(del_pkns)
        return add_pkd, del_pkd, uninstall_order

    def _check_configuration_hash(self, pkn):
        '''
        We want to see if the configuration that a given package has used
        has changed since it was installed.
        pkn -- the name of the package to check
        '''
        pkg = self._get_new_pkg(pkn)
        pkg_config = pkg.get_configuration()
        config_hash_path = os.path.join(pkg.get_path(), HASH_FILE)
        config_diff = self.config.check_hash(config_hash_path)
        differences = find_differences(pkg_config, config_diff, [])
        return differences

    def _uninstall_pkgs(self, del_pkd, uninstall_order,
                            dry_run=False):
        '''
        Given a dictionary of packages and an uninstall order, make some
        changes on this machine
        del_pkd -- A dictionary of package objects that are to
                   be uninstalled
        uninstall_order -- the order with which to uninstall packages
        dry_run -- boolean flag indicating if we're serious about this or not
        '''
        status = OK
        remove_full_pkns = []
        #Logger.info("UninstallOrder: %s" % uninstall_order)
        for name in uninstall_order:
            if del_pkd[name].full_name:
                name_str = del_pkd[name].full_name
            else:
                name_str = name
            remove_full_pkns.append(name_str)
        Logger.info("Packages to remove: %s" % remove_full_pkns)
        for pkn in uninstall_order:
            uninstall_status = del_pkd[pkn].uninstall(dry_run)
            if uninstall_status == FAIL:
                return FAIL
        return status

    def check_system(self):
        '''
        After a system has had all of its packages installed and it's in
        stable-state, this is a method we can use to verify that everything
        is still kosher
        '''
        Logger.info("System-check starting...")
        bom_pkns = self.config.get_bom_pkns()
        pdat = get_progress_data(self.instance_name)
        full_pdat = get_progress_data(self.instance_name, False)
        full_installed_pkns, _full_bpkns = get_installed(full_pdat)
        msg = "Packages that are installed: %s"
        Logger.info(msg % ' '.join(full_installed_pkns))
        installed_pkns, broken_pkns = get_installed(pdat)
        should_be_installed, shouldnt_be_installed = self._check_bom(bom_pkns)
        # check the configuration for each installed package
        pkg_info = {"ok": installed_pkns,
                    "reconfigure": {},
                    "not-installed": should_be_installed,
                    "remove": shouldnt_be_installed,
                    "broken": broken_pkns,
                   }
        for pkn in shouldnt_be_installed:
            pkg_info["ok"].remove(pkn)
        for pkn in installed_pkns:
            differences = self._check_configuration_hash(pkn)
            if differences:
                if pkn in pkg_info["ok"]:
                    pkg_info["reconfigure"][pkn] = differences
                    pkg_info["ok"].remove(pkn)
        for key in pkg_info:
            Logger.info("%s: %s" % (key, pkg_info[key]))
        return pkg_info

    def use_pkg(self, pkn, action, script_name=''):
        '''
        Main entry point to the class. Performs an action using a package
        pkn -- name of the package to use
        action -- STATUS, INSTALL, UNINSTALL, CONFIGURE, VERIFY, or EXEC
        script_name -- the name of a method to run within the package
        '''
        try:
            pkg = self._get_new_pkg(pkn)
            if pkg.status == FAIL:
                self.operation_status = FAIL
                return self._cleanup()
            if action == INSTALL:
                pdat = get_progress_data(self.instance_name, False)
                installed_pkns, broken_pkns = get_installed(pdat)
                if pkn in [installed_pkns + broken_pkns]:
                    Logger.error("Package %s cannot be installed." % pkn)
                    self.operation_status = FAIL
                    return FAIL
                add_pkd = {pkn:pkg}
                status = self._install_pkgs(add_pkd)
            if action == UNINSTALL:
                pdat = get_progress_data(self.instance_name, False)
                installed_pkns, broken_pkns = get_installed(pdat)
                bom_pkns = installed_pkns
                if pkn in bom_pkns:
                    bom_pkns.remove(pkn)
                self.config.set_bom_pkgs(bom_pkns)
                add_pkd, del_pkd, uninstall_order = self._ck_inst_stat([])
                status = self._uninstall_pkgs(del_pkd, uninstall_order)
            if action == VERIFY:
                status = pkg.verify()
            if action == CONFIGURE:
                self._check_configuration(pkg)
                status = pkg.configure()
                hash_path = os.path.join(pkg.get_path(), HASH_FILE)
                msg = "Writing configuration fingerprint to %s" % hash_path
                Logger.info(msg)
                self.config.save_hash(hash_path)
            if action == EXECUTE:
                status = pkg.execute_maint_script(script_name)
                if status == FAIL:
                    self.operation_status = FAIL
            msg = "Finished %s for %s."
            msg = msg % (ACTION_REVERSE_LOOKUP[action], pkn)
            self.operation_output.append(msg)
            status = self._cleanup()
            return self._cleanup()
        except Exceptions.BadPackage, err:
            errmsg = "Cannot perform action %s on package %s: %s"
            errmsg = errmsg % (ACTION_REVERSE_LOOKUP[action], err.pkn,
                               err.errmsg)
            Logger.warning(errmsg)
            Logger.info("RETURN FAIL 1")
            return FAIL

    def reconcile_system(self, action):
        '''
        Figure out everything that needs to happen to get this machine into
        compliance with its stated configuration and do it
        action -- DRY_RUN or RECONCILE
        '''

        self.operation_status = OK
        self.operation_output = []
        pkns = []
        dry_run = False

        if action == DRY_RUN:
            Logger.info("Reconcile dry-run starting...")
            dry_run = True
        else:
            Logger.info("Reconcile starting...")

        add_pkd, del_pkd, uninstall_order = self._ck_inst_stat(pkns)
        #Logger.info("uninstall_order: %s" % uninstall_order)
        status = self._uninstall_pkgs(del_pkd, uninstall_order,
                                          dry_run)
        if status == FAIL:
            msg = "Uninstallation failed. Aborting reconcile."
            self.operation_output.append(msg)
            self.operation_status = FAIL
            return self._cleanup()

        add_pkd, del_pkd, uninstall_order = self._ck_inst_stat(pkns)
        status = self._install_pkgs(add_pkd, dry_run)
        if status != OK:
            self.operation_status = FAIL
        self.operation_output.append("Finished installing")
        return self._cleanup()




#!/usr/bin/python

"Responsible for preparing filesystem for a package's use"

# BSD License
# Copyright (c) 2009, Peter Banka et al
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the GE Security nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from bombardier_core.static_data import OK, FAIL
from bombardier_core.static_data import VALID_PACKAGE_VERSIONS
import os, tarfile, sys, gzip, tarfile
import MetaData
from bombardier_core.mini_utility import get_package_path
from bombardier_core.mini_utility import get_spkg_path
import Exceptions
from bombardier_core.Logger import Logger
from commands import getstatusoutput as gso

BLOCK_SIZE = 10000


class Repository:
    '''Provides support functions for package, dealing with proper
    package file layout, unpacking, and verification'''

    def __init__(self, instance_name, pkg_data = {}):
        '''
        instance_name -- name of this machine
        pkg_data -- data regarding all packages on this machine
        '''
        self.instance_name = instance_name
        self.pkg_data      = pkg_data
        #self.packages      = pkg_data.keys()

    def get_meta_data(self, name):
        '''
        Query DSL database for this machine and return metadata
        for a given package name
        name -- name of the package
        '''
        if not name in self.pkg_data:
            msg = "Not found in %s" % self.pkg_data.keys()
            raise Exceptions.BadPackage(name, msg)
        data = self.pkg_data.get(name)
        return MetaData.MetaData(data)

    def unzip_type_4(self, pkg_path, full_name):
        '''
        Perform untar/ungzip operations. (NOTE: NOT RELIABLE IN Python)
        pkg_path -- full path to package file, minus extension
        full_name -- name of the package, including version
        '''
        Logger.info("Unzipping %s" % full_name)
        gzip_file = gzip.open(pkg_path + ".spkg")
        output_file = open(pkg_path + ".tar", 'wb')
        data = '1'
        while data:
            try:
                data = gzip_file.read(BLOCK_SIZE)
            except IOError, err:
                msg = "Error Reading %s: %s" % (full_name, err.__str__())
                Logger.error(msg)
                return FAIL
            except Exception, err:
                msg = "Corrupt package: %s (%s)" % (full_name, err.__str__())
                Logger.error(msg)
                return FAIL
            output_file.write(data)
        output_file.close()
        gzip_file.close()
        return OK

    def get_type_4(self, full_name):
        '''
        Get a type-4 package from the filesystem, and process it
        full_name -- name of package (with version)
        '''
        pkg_dir = get_package_path(self.instance_name)
        pkg_path = os.path.join(pkg_dir, full_name)
        if not os.path.isfile(pkg_path + ".spkg"):
            erstr = "No package file in %s." % (pkg_path + ".spkg")
            Logger.error(erstr)
            raise Exceptions.BadPackage(full_name, erstr)
        if sys.platform != 'win32':
            cmd = "cd %s && tar -xzf %s.spkg" % (pkg_dir, full_name)
            Logger.info("Untarring with command: %s" %cmd)
            if not os.system(cmd) == OK:
                raise Exceptions.BadPackage(full_name, "Could not unpack")
            return OK
        if self.unzip_type_4(pkg_path, full_name) == FAIL:
            raise Exceptions.BadPackage(full_name, "could not unzip")
        tar = tarfile.open(pkg_path + ".tar", "r")
        tar.errorlevel = 2
        cwd = os.getcwd()
        os.chdir(pkg_dir)
        for tarinfo in tar:
            try:
                tar.extract(tarinfo)
            except tarfile.ExtractError, err:
                Logger.warning("Error with package %s,%s: "\
                               "%s" % (full_name, tarinfo.name, err))
        tar.close()
        if not os.path.isdir(os.path.join(pkg_path, full_name)):
            erstr = "Package %s is malformed." % (full_name)
            os.chdir(cwd)
            raise Exceptions.BadPackage(full_name, erstr)
        os.chdir(cwd)
        os.unlink(pkg_path + ".tar")
        return OK

    def hunt_and_explode(self):
        '''
        Used to find all type-5 package data in the repository and
        untar the files properly.
        '''
        _status, output = gso('find /opt/spkg/ -name "*.tar.gz"')
        start_dir = os.getcwd()
        for full_tar_file_name in output.split('\n'):
            tmp_list = full_tar_file_name.split(os.path.sep)
            tar_file_name = tmp_list[-1]
            base_name = tar_file_name.split('.tar.gz')[0]
            tar_file_dir = os.path.sep.join(tmp_list[:-1] + [base_name])
            if not os.path.isdir(tar_file_dir):
                Logger.info("Exploding %s..." % base_name)
                cmd = "mkdir -p %s" % tar_file_dir
                status = os.system(cmd)
                if status == OK:
                    cmd = "cd %s && tar -xzf ../%s"
                    cmd = cmd % (tar_file_dir, tar_file_name)
                    if os.system(cmd) != OK:
                        msg = "could not untar %s" % (tar_file_name)
                        raise Exceptions.BadPackage(full_name, msg)
        os.chdir(start_dir)

    def make_symlinks(self, pkg_path, info_dict, full_name):
        '''
        Create symlinks from the repository into where the package
        expects to run
        pkg_path -- directory where the versioned package will reside
        info_dict -- information regarding injectors and libraries
        full_name -- name of the package, including version info
        '''
        base_path = os.path.join(get_spkg_path(), "repos")
        for component_type in info_dict:
            #Logger.info("Component: %s" % component_type)
            #Logger.info("info_dict: %s" % info_dict)
            component_dict = info_dict[component_type]
            for component_name in component_dict:
                full_path = component_dict[component_name]["path"]
                full_name = full_path.split(os.path.sep)[-1]
                full_name = full_name.split('.tar.gz')[0]
                src = os.path.join(base_path, component_type, full_name)
                dst = os.path.join(pkg_path, component_type, component_name)
                if os.path.islink(dst):
                    continue
                cmd = "ln -s %s %s" % (src, dst)
                #Logger.info("CMD: (%s)" % cmd)
                if os.system(cmd) != OK:
                    msg = "Could not create symlink (%s)" % cmd
                    raise Exceptions.BadPackage(full_name, msg)

    def get_type_5(self, full_name, injectors_info, libs_info):
        '''
        Get type-5 package components from the filesystem, and process them
        full_name -- name of the package, including version info
        injectors_info -- dictionary describing injector libraries
        libs_info -- dictionary describing python code libraries
        '''
        self.hunt_and_explode()
        pkg_path = os.path.join(get_spkg_path(), self.instance_name,
                                "packages", full_name)
        injector_path = os.path.join(pkg_path, "injectors")
        lib_path = os.path.join(pkg_path, "libs")
        if not os.path.isdir(pkg_path):
            Logger.info("Making directory %s" % pkg_path)
            for path in [pkg_path, injector_path, lib_path]:
                cmd = "mkdir -p %s" % path
                if os.system(cmd) != OK:
                    msg = "Could not create directory structure (%s)" % path
                    raise Exceptions.BadPackage(full_name, msg)
        info_dict = {"injectors": injectors_info,
                     "libs": libs_info}
        self.make_symlinks(pkg_path, info_dict, full_name)
        return OK

    def determine_pkg_version(self, pkn):
        'We need to know what version a package is'
        meta_data = self.get_meta_data(pkn)
        version = meta_data.data.get("package-version")
        if type(version) == type(1):
            if version in VALID_PACKAGE_VERSIONS:
                return version
        msg = "Unknown package version"
        raise Exceptions.BadPackage, (pkn, msg)




"""This is little class that acts like both a dictionary
and a ConfigParser object for convenience sake."""

# Copyright (C) 2005 Peter Banka

import ConfigParser

class MetaData:
    "Class that provides ConfigParser and dict interface"
    def __init__(self, data):
        "data -- dictionary the class will use"
        self.data = data

    def __getitem__(self, key):
        "part of dict interface"
        return self.data[key]

    def __setitem__(self, key, value):
        "part of dict interface"
        self.data[key] = value

    def has_key(self, key):
        "part of dict interface"
        return self.data.has_key(key)

    def get(self, section_name, option_name, default=None):
        "part of ConfigParser interface"
        if not self.data.has_key(section_name):
            if default != None:
                return default
            raise ConfigParser.NoSectionError(section_name)
        section = self.data[section_name]
        if not section.has_key(option_name):
            if default != None:
                return default
            raise ConfigParser.NoOptionError(section_name, option_name)
        return section[option_name]

    def has_section(self, section_name):
        "part of ConfigParser interface"
        if not self.data.has_key(section_name):
            if section_name in self.data.keys():
                return True
            return False




#!/usr/bin/python

"Implements a single-file package, subclassing the Package class"

# BSD License
# Copyright (c) 2009, Peter Banka et al
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the GE Security nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os, glob, random
import sys

from bombardier_core import Spkg
from bombardier_core.mini_utility import get_package_path
from Exceptions import BadPackage, FeatureRemovedException
from Job import Job
from Exceptions import RebootRequiredException
from bombardier_core.Logger import Logger
from bombardier_core.static_data import OK, FAIL, REBOOT
from bombardier_core.static_data import INSTALL, UNINSTALL
from bombardier_core.static_data import CONFIGURE, VERIFY

from Package import Package

def eval_boolean(data):
    """Someone has typed in some data in a configuration file that should
    be treated as boolean data. Convert it."""
    if not data:
        return False
    if type(data) != type('string'):
        if type(data) == type(1):
            if data == 1:
                return True
            else:
                return False
        else:
            return False
    data = data.strip().upper()
    if data in ["TRUE", "YES", "1", "OK"]:
        return True
    return False


class PackageV4(Package):

    """This class provides an abstraction for a downloadable,
    installable, verifiable, and uninstallable thing. Each
    package is held in the repository"""

    def __init__(self, name, repository, config, 
                 instance_name):
        '''
        name -- the name of this package
        repository -- object that keeps track of package data
        config -- configuration for this machine
        instance_name -- name of this machine
        '''
        Package.__init__(self, name, repository, config,
                         instance_name)
        self.package_version = 4

    ############################################ PUBLIC METHODS

    def initialize(self):
        '''Evaluate the package meta-data for consistency
        and initialize parameters. Verify that data on the disk for
        this package is consistent.
        '''
        Package.initialize(self)
        self.status = FAIL
        if self.meta_data.data:
            if type(self.meta_data.data) == type(dict()):
                install_data = self.meta_data.data.get("install")
                chk = install_data.get('console')
                self.console = eval_boolean(chk)
                chk = install_data.get('reboot')
                self.reboot = eval_boolean(chk)
                self.checksum = install_data.get('md5sum')
                if install_data:
                    if type(install_data) == type({}):
                        self.full_name = install_data.get('fullName')
                        if not self.full_name:
                            self.full_name = install_data.get('full_name')
                        if self.full_name:
                            self.status = OK
                            return
                        else:
                            msg = "Package does not exist on the server"
                            raise BadPackage, (self.name, msg)
                    else:
                        msg = "The server's record is completely corrupt"
                        raise BadPackage, (self.name, msg)
                else:
                    msg = "server record is corrupt: no 'install' section"
                    raise BadPackage, (self.name, msg)
            else:
                msg = "server record is corrupt: not a dictionary"
                raise BadPackage, (self.name, msg)
        else:
            msg = "No metadata found for this package"
            raise BadPackage, (self.name, msg)

    def execute_maint_script(self, script_name):
        '''
        execute a user-defined function
        script_name -- name of the function to run
        '''
        Package.execute_maint_script(self)
        script_path = "%s/%s.py" % (self.maint_dir, script_name)
        start_dir = os.getcwd()
        os.chdir(self.maint_dir)
        if not os.path.isfile(script_path):
            msg = "%s does not exist" % script_path
            raise BadPackage, (self.name, msg)
        sys.path.append(self.maint_dir )
        import_string = 'import %s' % script_name
        cmd = 'status = %s.execute(self.config, Logger)' % script_name
        job = Job(import_string, cmd, self.config)
        status = job.execute()
        sys.path.remove(self.maint_dir)
        os.chdir(start_dir)
        if status == None:
            status = OK
        if type(status) != type(1):
            msg = "Invalid status type (%s: '%s')"
            Logger.warning(msg % (type(status), status))
            status = FAIL
        return status

    ############################################ PRIVATE METHODS

    def _eval_priority(self):
        'determine priority of this package'
        self.priority = self.meta_data.data["install"].get('priority')
        Package._eval_priority(self)

    def _download(self):
        '''
        Run right before a package action takes place.
        We see if we have a package directory and if not, tell the
        repository to unpack it from the filesystem. Then verify all
        the pieces and parts are there for a good type-4 package
        '''
        if not self.downloaded:
            pkg_dir = os.path.join(get_package_path(self.instance_name),
                                                  self.full_name)
            if not os.path.isdir(pkg_dir):
                self.repository.get_type_4(self.full_name)
            self.scripts_dir = os.path.join(pkg_dir, "scripts")
            self.maint_dir = os.path.join(pkg_dir, "maint")
            injector_dir = os.path.join(pkg_dir, "injector")
            for required_dir in [self.scripts_dir, injector_dir]:
                if not os.path.isdir(required_dir):
                    errmsg = "Required directory %s does not exist"
                    errmsg = errmsg % required_dir
                    self.status = FAIL
                    raise BadPackage(self.name, errmsg)
            self.working_dir = injector_dir
            self.downloaded = True

    def _get_possible_module_files(self):
        '''In type-4 packages, we play a guessing game which python script
        to use as the one to run the installation'''
        files = glob.glob("*.py")
        files = [x.split('.py')[0] for x in files]
        if self.name in files:
            files = [self.name]
        else:
            vpkg_name = self.meta_data.data.get('virtualpackage')
            if vpkg_name in files:
                files = [vpkg_name]
        return files

    def _find_cmd(self, action, future_pkns, dry_run):
        '''
        Perform the action on the system, importing modules from the package
        and running the appropriate method on the class within.
        action -- INSTALL, UNINSTALL, CONFIGURE, VERIFY
        future_pkns -- future package names. Some packages want to know
                       about the packages that will come after them
        dry_run -- boolean flag to see if we're really going to do this
        '''
        cwd = os.getcwd()
        sys.path.insert(0, self.scripts_dir)
        os.chdir(self.scripts_dir)
        files = self._get_possible_module_files()
        status = FAIL
        file_found = False
        for file_name in files:
            try:
                obj = Spkg.SpkgV4(self.config, logger=Logger)
                os.chdir(self.working_dir)
                letters = [ chr( x ) for x in range(65, 91) ]
                random.shuffle(letters)
                rand_string = ''.join(letters)
                exec("import %s as %s" % (file_name, rand_string))
                self.config["__FUTURE_PACKAGES__"] = future_pkns
                self.config["__INSTANCE__"] = self.instance_name
                cmd_str = "obj = %s.%s(self.config, Logger)"
                exec(cmd_str % (rand_string, file_name))
                file_found = True
                if not dry_run:
                    if action == INSTALL:
                        status = obj.installer()
                    elif action == VERIFY:
                        status = obj.verify()
                    elif action == UNINSTALL:
                        status = obj.uninstaller()
                    elif action == CONFIGURE:
                        status = obj.configure()
                    else:
                        raise FeatureRemovedException(action)
                else:
                    status = OK
                del rand_string
                if file_name in sys.modules:
                    sys.modules.pop(file_name)
                while self.scripts_dir in sys.path:
                    sys.path.remove(self.scripts_dir)
                break
            except ImportError:
                msg = "File %s is not runnable. Looking for others" % file_name
                Logger.debug(msg)
                continue
            except SystemExit, err:
                if err.code:
                    status = err.code
                else:
                    status = 0
                file_found = True
                break
            except KeyboardInterrupt:
                Logger.error("Keyboard interrupt detected. Exiting...")
                sys.exit(10)
            except SyntaxError, err:
                self._dump_error(err, file_name)
                file_found = False
                break
            except StandardError, err:
                self._dump_error(err, file_name)
                file_found = True
                status = FAIL
                break
        os.chdir(cwd)
        if not file_found:
            msg = "Unable to find a suitable script to install."
            raise BadPackage(self.name, msg)
        if status == None:
            status = OK
        if status == REBOOT:
            raise RebootRequiredException(self.name)
        if status != OK:
            erstr = "%s: failed with status %s" % (self.full_name, status)
            Logger.error(erstr)
            return FAIL
        return OK




#!/usr/bin/env python

"sets bombardier_client up."

from distutils.core import Command, setup
from unittest import TextTestRunner, TestLoader
from glob import glob
from os.path import splitext, basename, join as pjoin, walk
import os
from lib._version import version_info

LONG_DESCRIPTION = """\
Bombardier is a software system that delivers visibility, control
and automation to datacenter environments. Bombardier provides a means
for changes to be rolled out to a network of servers in a highly
controlled way, providing optimum security, logging, centralized
change control.

This package is called the "Bombardier Client," and its purpose is to
provide some facilities for the Bombardier Server to make it easier
for the server to do its job.  If you're a stickler for terminology,
then, it's not really a client at all, since is not used to connect to
a server.
"""


class CleanCommand(Command):
    user_options = [ ]
    def initialize_options(self):
        self._clean_me = [ ]
        for root, dirs, files in os.walk('.'):
            for f in files:
                if f.endswith('.pyc'):
                    self._clean_me.append(pjoin(root, f))

    def finalize_options(self):
        pass

    def run(self):
        for clean_me in self._clean_me:
            try:
                os.unlink(clean_me)
            except:
                pass

class TestCommand(Command):
    user_options = [ ]
    def initialize_options(self):
        self._dir = os.getcwd()

    def finalize_options(self):
        pass

    def run(self):
        '''
        Finds all the tests modules in tests/, and runs them.
        '''
        testfiles = [ ]
        for t in glob(pjoin(self._dir, 'tests', '*.py')):
            if not t.endswith('__init__.py'):
                testfiles.append('.'.join(
                    ['tests', splitext(basename(t))[0]])
                )

        tests = TestLoader().loadTestsFromNames(testfiles)
        t = TextTestRunner(verbosity = 1)
        t.run(tests)

setup (name = "bombardier_client",
       description = "Open Source Configuration management and package delivery: machine component",
       version = str("%(branch_nick)s-%(revno)d" % version_info),
       long_description = LONG_DESCRIPTION,
       packages = ["bombardier_client"],
       package_dir = {"bombardier_client": "lib"},
       scripts = ["scripts/bc.py"],
       author = "Peter Banka and Shawn Sherwood",
       author_email = "peter.banka@gmail.com",
       url = "http://bombardierinstaller.org/",
       install_requires = [ 'PyYAML', 'bombardier_core' ],
       provides = [ 'bombardier_client' ],
       classifiers = [
           "Development Status :: 2 - Pre-Alpha",
           "Programming Language :: Python",
           "Intended Audience :: Developers",
           "Intended Audience :: Information Technology",
           "License :: OSI Approved :: BSD License",
           "Operating System :: POSIX",
           "Topic :: Software Development :: Libraries :: Python Modules",
           "Topic :: System :: Archiving :: Packaging",
       ],
       cmdclass = { 'test': TestCommand, 'clean': CleanCommand }
)



#!/usr/bin/python

'Runs a package job in an isolated thread'

# BSD License
# Copyright (c) 2009, Peter Banka et al
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the GE Security nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import StringIO, traceback, time
from bombardier_core.Logger import Logger
from bombardier_core.static_data import OK, FAIL
from threading import Thread

class JobThread(Thread):
    'Runs an action on a machine in a separate thread'

    def __init__(self, import_string, cmd, config):
        '''
        import_string -- python code which will be exec'd to import the
                         package class
        cmd -- python code which will be exec'd which will run the actual
               action on this machine.
        config -- the configuration of this machine
        '''
        Thread.__init__(self)
        self.import_string = import_string
        self.cmd = cmd
        self.config = config
        self.cmd_status = None

    def run(self):
        'Thread interface'
        Logger.debug("Running %s..." % self.cmd)
        try:
            exec(self.import_string)
            exec("self.cmd_status = %s" % self.cmd)
        except StandardError, err:
            Logger.error("Failed to run %s (%s)" % (self.cmd, err))
            sio = StringIO.StringIO()
            traceback.print_exc(file=sio)
            sio.seek(0)
            data = sio.read()
            ermsg = ''
            for line in data.split('\n'):
                ermsg += "\n||>>>%s" % line
            Logger.error(ermsg)
            self.cmd_status = FAIL

class Job:
    'Controls the execution of a thread'

    def __init__(self, import_string, cmd, config):
        '''
        import_string -- python code which will be exec'd to import the
                         package class
        cmd -- python code which will be exec'd which will run the actual
               action on this machine.
        config -- the configuration of this machine
        '''
        self.import_string = import_string
        self.cmd = cmd
        self.config = config
        self.start_time = None
        self.job_thread = None
        self.job_status = None

    def is_running(self):
        'ability to check if the job is finished'
        if self.job_thread:
            if self.job_thread.isAlive():
                return True
            if type(self.job_thread.cmd_status) == type(0) \
               or type(self.job_thread.cmd_status) == type('str'):
                Logger.debug("-- status: %s" % (self.job_thread.cmd_status))
                self.job_status = self.job_thread.cmd_status
            else:
                msg = "Invalid return status (type: %s)"
                Logger.error(msg % (type(self.job_thread.cmd_status)))
                self.job_status = FAIL
        return False

    def execute(self):
        'starts a job and waits for it to finish'
        self.start_time = time.time()
        status = FAIL
        if self.is_running():
            msg = "Refusing to run %s; it is already running" % self.cmd
            Logger.error(msg)
            return FAIL
        self.job_thread = JobThread(self.import_string, self.cmd, self.config)
        self.job_thread.start()
        Logger.info("Started job...")
        counter = 1
        while self.is_running():
            time.sleep(1)
            counter += 1
            if not counter % 100:
                msg = "Waiting for completion (%s)..."
                Logger.info(msg % time.strftime(time.ctime()))
                counter = 1
        status = self.job_thread.cmd_status
        return status

    def kill_thread(self, timeout=5):
        'Aborts a job'
        if not self.is_running():
            return OK
        self.job_thread.join(timeout)
        return OK





#!/usr/bin/python

'Abstract class that implements a package'

# BSD License
# Copyright (c) 2009, Peter Banka et al
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the GE Security nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os, time
import StringIO, traceback

import MetaData
from bombardier_core.mini_utility import get_package_path, get_spkg_path
from bombardier_core.mini_utility import update_progress, get_progress_data

from Exceptions import BadPackage, FeatureRemovedException
from bombardier_core.Logger import Logger
from bombardier_core.static_data import OK, FAIL
from bombardier_core.static_data import INSTALL, UNINSTALL, CONFIGURE, VERIFY

AVERAGE = 100

class Package:

    """This class provides an abstraction for a downloadable,
    installable, verifiable, and uninstallable thing. Each
    package is held in the repository"""

    def __init__(self, name, repository, config, instance_name):
        '''
        name -- name of the package
        repository -- object that manages files on the disk for the package
        config -- configuration of the machine
        instance_name -- name of this machine
        '''
        self.name         = name
        self.repository   = repository
        self.instance_name = instance_name
        self.action       = INSTALL
        self.installed    = False
        self.dependencies = []
        self.depends_on_me = []
        self.console      = False
        self.reboot       = False
        self.package_version = 0
        self.checksum     = ''
        self.config       = config
        self.status       = OK
        self.priority     = AVERAGE
        self.meta_data    = MetaData.MetaData({})
        self.working_dir  = ''
        self.scripts_dir  = ''
        self.maint_dir    = ''
        self.downloaded   = False
        self.dependency_errors = []
        self.full_name    = None

    ############################################ PUBLIC METHODS

    def initialize(self):
        '''Evaluate the package meta-data for consistency
        and initialize parameters. Verify that data on the disk for
        this package is consistent.
        '''
        try:
            self.meta_data = self.repository.get_meta_data(self.name)
        except BadPackage, bpe:
            Logger.error("initialize INVALIDATING")
            self._invalidate(bpe)
            return
        self._gather_dependencies()
        self._eval_priority()

    def get_configuration(self):
        'getter method'
        return self.meta_data.data.get("configuration", {})

    def get_priority(self):
        'getter method'
        return self.priority

    def add_dependency_error(self, dependency_name):
        '''
        Record the fact that the system has a dependency error.
        This is used when uninstalling packages later.
        dependency_name -- name of the dependency that is missing
        '''
        errmsg = "BOM file is incomplete: should contain %s" % dependency_name
        Logger.warning(errmsg)
        self.dependency_errors.append(dependency_name)

    def install_and_verify(self, pkns, dry_run=False):
        '''
        Perform an installation and verification operation
        pkns -- list of package names
        dry_run -- whether to actually perform the action
        '''
        self._download()
        if self.action == INSTALL:
            status = self._install(pkns, dry_run=dry_run)
            if not dry_run:
                if status == OK:
                    status = self.verify()
                self._write_progress()
            return status
        raise FeatureRemovedException(self.action)

    def configure(self, dry_run=False):
        '''
        Perform a configuration action on this package
        dry_run -- whether to actually perform the action
        '''
        self._download()
        return self._find_cmd(CONFIGURE, [], dry_run)

    def verify(self, dry_run=False): 
        '''
        Perform a verification action on this package
        dry_run -- whether to actually perform the action
        '''
        self._download()
        message = "Verifying package %s" % self.full_name
        Logger.info(message)
        self.status = self._find_cmd(VERIFY, [], dry_run)
        if self.action != INSTALL:
            self._write_progress()
        Logger.info("Verify result for %s : %s" % (self.full_name, self.status))
        return self.status

    def uninstall(self, dry_run=False):
        '''
        Perform an uninstall action on this package
        dry_run -- whether to actually perform the action
        '''
        self.action = UNINSTALL
        self._download()
        if self.status != OK:
            return self.status
        dry_run_string = ""
        if dry_run:
            dry_run_string = " --DRY_RUN-- "
        Logger.info("Uninstalling package %s%s" % (self.name, dry_run_string))
        self.status = self._find_cmd(UNINSTALL, [], dry_run)
        if not dry_run:
            self._write_progress()
        msg = "Uninstall result for %s%s : %s"
        Logger.info(msg % (self.full_name, dry_run_string, self.status))
        return self.status

    def execute_maint_script(self, script_name):
        '''
        execute a user-defined function
        script_name -- name of the function to run
        '''
        self._download()
        # remove old history
        output_path = os.path.join(get_spkg_path(), self.instance_name,
                                  "output", "%s-output.yml" % script_name)
        if os.path.isfile(output_path):
            os.unlink(output_path)
        message = "Executing (%s) inside package (%s)"
        Logger.info(message % (script_name, self.full_name))

    def get_path(self):
        'find place on the disk where this package can be accessed'
        path = os.path.join(get_package_path(self.instance_name), 
                            self.full_name)
        return path

    ############################################ PRIVATE METHODS

    def _download(self):
        'Virtual method'
        pass

    def _find_cmd(self, action, pkns, dry_run):
        'Virtual method'
        return FAIL

    def _invalidate(self, exception_object):
        'set the package as not operable'
        erstr = "INVALID PACKAGE: %s" % self.name
        Logger.error(erstr)
        Logger.error(str(exception_object))
        self.status = FAIL

    def _eval_priority(self):
        'determine priority of this package'
        if not self.priority:
            self.priority = AVERAGE
        else:
            try:
                self.priority = int(self.priority)
            except ValueError:
                Logger.warning(ermsg)
                self.priority = AVERAGE
        
    def _gather_dependencies(self):
        'determine what packages must be installed before this package'
        self.dependencies = self.meta_data.data.get("dependencies")
        if type(self.dependencies) == type(None):
            self.dependencies = []
        elif type(self.dependencies) == type('string'):
            self.dependencies = [self.dependencies]
        elif type(self.dependencies) == type([]):
            pass
        else:
            errmsg = "Unsupported dependencies data for %s. "\
                     "Must be a list. [%s] instead."
            Logger.warning(errmsg % (self.name, self.dependencies))
            self.dependencies = []
    
    @classmethod
    def _dump_error(cls, err, file_name):
        '''
        perform a traceback for logging purposes
        err -- an exception object
        file_name -- the module that was running
        '''
        Logger.error("Error detected in %s (%s)." % (file_name, err))
        sio = StringIO.StringIO()
        traceback.print_exc(file=sio)
        sio.seek(0)
        data = sio.read()
        ermsg = ''
        for line in data.split('\n'):
            ermsg += "\n||>>>%s" % line
        Logger.error(ermsg)
        Logger.error("Error ocurred in %s" % file_name)

    def _install(self, pkns, dry_run=False):
        '''
        perform an installation operation
        pkns -- package names to be installed after this one
        dry_run -- whether or not to actuall perform the installation
        '''
        dry_run_string = ""
        if dry_run:
            dry_run_string = " --DRY_RUN-- "
        self._download()
        message = "Beginning installation of (%s)%s"
        Logger.info(message % (self.full_name, dry_run_string))
        self.status = self._find_cmd(INSTALL, pkns, dry_run)
        msg = "Install result for %s%s : %s"
        Logger.info(msg % (self.full_name, dry_run_string, self.status))
        return self.status

    def _write_install(self, pdat, time_string):
        if pdat.get(self.full_name) == None:
            pdat[self.full_name] = {}
        if self.full_name:
            if self.status == OK:
                pdat[self.full_name]['INSTALLED']   = time_string
                pdat[self.full_name]['VERIFIED']    = time_string 
                pdat[self.full_name]['UNINSTALLED'] = 'NA'

            else:
                pdat[self.full_name]['INSTALLED']   = "BROKEN"
                pdat[self.full_name]['VERIFIED']    = 'NA'
                pdat[self.full_name]['UNINSTALLED'] = 'NA'
        else:
            msg = "Package installed that doesn't have a full_name"
            raise BadPackage(self.name, msg)
        return pdat

    def _write_uninstall(self, pdat, time_string):
        like_packages = []
        for pkn in pdat:
            if pkn.startswith(self.name) and pkn != self.full_name:
                like_packages.append(pkn)
        for pkn in like_packages:
            del pdat[pkn]
        if self.status == OK:
            pdat[self.full_name]['UNINSTALLED'] = time_string
            pdat[self.full_name]['INSTALLED']   = 'NA'
        else:
            pdat[self.full_name]['UNINSTALLED'] = "BROKEN"
        return pdat

    def _write_progress(self):
        '''Write out information about the state of the packages on this
        system. Keeps track of when packages were installed, uninstalled,
        verified, and keeps track of dependency errors.'''
        time_string = time.ctime()
        pdat = get_progress_data(self.instance_name)
        if not pdat.has_key(self.full_name):
            pdat[self.full_name] = {"INSTALLED": "NA",
                                             "UNINSTALLED": "NA",
                                             "VERIFIED": "NA",
                                             "DEPENDENCY_ERRORS": []}
        if self.action == INSTALL:
            pdat = self._write_install(pdat, time_string)
    
        elif self.action == UNINSTALL:
            pdat = self._write_uninstall(pdat, time_string)
            like_packages = []
            for pkn in pdat:
                if pkn.startswith(self.name) and pkn != self.full_name:
                    like_packages.append(pkn)
            for pkn in like_packages:
                del pdat[pkn]
            if self.status == OK:
                pdat[self.full_name]['UNINSTALLED'] = time_string
                pdat[self.full_name]['INSTALLED']   = 'NA'
            else:
                pdat[self.full_name]['UNINSTALLED'] = "BROKEN"

        elif self.action == VERIFY:
            pdat[self.full_name]['VERIFIED'] = time_string
        pdat[self.full_name]['DEPENDENCY_ERRORS'] = self.dependency_errors
        update_progress({"install-progress":pdat},
                        self.instance_name, overwrite=True)
        return OK



#!/usr/bin/python

"Implements a multi-file package, subclassing the Package class"

# BSD License
# Copyright (c) 2009, Peter Banka et al
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the GE Security nor the names of its contributors may
#   be used to endorse or promote products derived from this software without
#   specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os, sys, random

from bombardier_core import Spkg
import StringIO, traceback
from bombardier_core.mini_utility import get_spkg_path
from bombardier_core.mini_utility import get_package_path
from Exceptions import BadPackage
from Exceptions import RebootRequiredException
from bombardier_core.Logger import Logger
from bombardier_core.static_data import OK, FAIL, REBOOT
from bombardier_core.static_data import ACTION_REVERSE_LOOKUP
from Package import Package

class PackageV5(Package):

    """This class provides an abstraction for a downloadable,
    installable, verifiable, and uninstallable thing. Each
    package is held in the repository"""

    def __init__(self, name, repository, config,
                 instance_name):
        Package.__init__(self, name, repository, config,
                         instance_name)
        self.package_version = 5
        self.release = None
        self.injectors_info = {}
        self.libs_info = {}
        self.class_name = None

    ############################################ PUBLIC METHODS

    def initialize(self):
        '''Evaluate the package meta-data for consistency
        and initialize parameters. Verify that data on the disk for
        this package is consistent.
        '''
        Package.initialize(self)
        self.status = FAIL
        dat = self.meta_data.data
        if dat:
            release = dat.get('release')
            if type(release) == type(1):
                self.release = release
                class_name = dat.get('class_name')
                if type(class_name) == type('string'):
                    self.status = OK
                    self.class_name = class_name
                    injectors_info = dat.get('injectors')
                    if type(injectors_info) == type({}):
                        self.injectors_info = injectors_info
                    libs_info = dat.get('libs')
                    if type(libs_info) == type({}):
                        self.libs_info = libs_info
                else:
                    msg = "No script infomation for this package"
                    raise BadPackage, (self.name, msg)
            else:
                msg = "No release infomation for this package"
                raise BadPackage, (self.name, msg)
        else:
            msg = "No metadata found for this package"
            raise BadPackage, (self.name, msg)
        self.full_name = "%s-%d" % (self.name, self.release)

    def execute_maint_script(self, script_name):
        '''
        execute a user-defined function
        script_name -- name of the function to run
        '''
        Package.execute_maint_script(self, script_name)
        self.status = self._find_cmd(script_name, [], False)
        msg = "%s result for %s : %s"
        Logger.info(msg % (script_name, self.full_name, self.status))
        return self.status

    ############################################ PRIVATE METHODS

    def _eval_priority(self):
        'determine priority of this package'
        self.priority = self.meta_data.data.get('priority')
        Package._eval_priority(self)

    def _download(self):
        '''
        Run right before a package action takes place.
        We see if we have a package directory and if not, tell the
        repository to unpack it from the filesystem. Then verify all
        the pieces and parts are there for a good type-4 package
        '''
        if not self.downloaded:
            self.repository.get_type_5(self.full_name, self.injectors_info,
                                       self.libs_info)
            pkg_dir = os.path.join(get_package_path(self.instance_name),
                                                  self.full_name)
            injector_dir = os.path.join(pkg_dir, "injectors")
            if os.path.isdir(injector_dir):
                self.working_dir = injector_dir
            else:
                self.status = FAIL
                errmsg = "The injector directory does not exist for [%s]"
                raise BadPackage(self.name, errmsg % self.full_name)
        self.downloaded = True

    def _get_lib_path(self):
        '''
        Need to modify our path and then clean it out again. This gets
        the data that both operations will need.
        '''
        package_path = os.path.join(get_spkg_path(), self.instance_name,
                                    "packages", self.full_name)
        lib_path = os.path.join(package_path, "libs")
        return lib_path

    def _get_object(self, future_pkns):
        '''
        Import modules from the package in order to run an appropriate method
        on the class within.
        future_pkns -- future package names. Some packages want to know
                       about the packages that will come after them
        '''
        lib_path = self._get_lib_path()
        sys.path.insert(0, lib_path)
        obj = None

        try:
            class_name = '.'.join(self.class_name.split('.')[1:])
            obj = Spkg.SpkgV5(self.config)
            os.chdir(self.working_dir)
            letters = [ chr( x ) for x in range(65, 91) ]
            random.shuffle(letters)
            rand_string = ''.join(letters)
            exec("import %s as %s" % (self.class_name, rand_string))
            self.config["__FUTURE_PACKAGES__"] = future_pkns
            self.config["__INSTANCE__"] = self.instance_name
            cmd_str = "obj = %s.%s(self.config)"
            exec(cmd_str % (rand_string, class_name))
        except ImportError, ie:
            tb_str = StringIO.StringIO()
            traceback.print_exc(file=tb_str)
            tb_str.seek(0)
            data = tb_str.read()
            ermsg = ''
            for line in data.split('\n'):
                Logger.error(line)
            msg = "Class %s is not importable." % self.class_name
            raise BadPackage(self.name, msg)
        return obj, rand_string

    def _cleanup(self, obj):
        '''
        Running a package action is messy business. Here's where we
        attempt to clean up the mess
        '''
        lib_path = self._get_lib_path()
        if self.class_name in sys.modules:
            sys.modules.pop(self.class_name)
        sys.path.remove(lib_path)

    def _find_cmd(self, action, future_pkns, dry_run):
        '''
        Perform the action on the system, importing modules from the package
        and running the appropriate method on the class within.
        action -- INSTALL, UNINSTALL, CONFIGURE, VERIFY
        future_pkns -- future package names. Some packages want to know
                       about the packages that will come after them
        dry_run -- boolean flag to see if we're really going to do this
        '''
        if type(action) == type(1):
            action=ACTION_REVERSE_LOOKUP[action]
        cwd = os.getcwd()
        obj, rand_string = self._get_object(future_pkns)
        try:
            if not hasattr(obj, action):
                msg = "Class %s does not have a %s method."
                raise BadPackage(self.name, msg % (self.class_name, action))
            if not dry_run:
                exec("status = obj.%s()" % action)
            else:
                status = OK
            self._cleanup(obj)
            del rand_string
        except SystemExit, err:
            if err.code:
                status = err.code
            else:
                status = OK
            del rand_string
        except KeyboardInterrupt:
            Logger.error("Keyboard interrupt detected. Exiting...")
            status = FAIL
            sys.exit(10) # FIXME: Literal
        except SyntaxError, err:
            self._dump_error(err, self.class_name)
            status = FAIL
            del rand_string
        except StandardError, err:
            self._dump_error(err, self.class_name)
            status = FAIL
            del rand_string
        os.chdir(cwd)
        if status == None:
            status = OK
        if status == REBOOT:
            raise RebootRequiredException(self.name)
        if status != OK:
            erstr = "%s: failed with status %s" % (self.full_name, status)
            Logger.error(erstr)
            return FAIL
        return OK

